import time
from raft.tasks import task
from ..base import AwsTask


@task(klass=AwsTask, help={
    'cluster': 'ecs cluster name',
    'service': 'ecs service name, must be running on `cluster`',
    'interval': 'the polling interval at which logs are checked, default is 10 seconds',
    'tail': 'the number of previous events to show, default is 10',
})
def watch_service_deploy(ctx, cluster, service, interval=10, tail=10, session=None, **kwargs):
    """
    watches the ecs service event log until a deployment is completed or fails.

    n.b., the timestamps on the service event log are displayed in
    US/Central timezone.
    """
    from botocore.exceptions import ClientError
    from pytz import timezone
    from ...base.utils import notice_end
    ecs = session.client('ecs')
    cst6cdt = timezone('US/Central')
    seen_ids = set()
    primary = None
    first = True
    while True:
        try:
            response = ecs.describe_services(cluster=cluster, services=[service])
        except ClientError as err:
            if err.response['Error']['Code'] == 'ServiceNotActiveException':
                notice_end(f'service {service} not active.')
                return
            # Anything else coming from ECS → propagate or handle as needed
            raise
        service0 = response['services'][0]
        deployments = service0.get('deployments') or []
        primary = None
        for x in deployments:
            if x['status'] == 'PRIMARY':
                primary = x
                break
        # ECS returns events in reverse chronological order (newest first)
        # so we reverse to print oldest first
        new_events = []
        for event in service0.get('events', []):
            st_id = event['id']
            if st_id not in seen_ids:
                new_events.append(event)
                seen_ids.add(st_id)
                if first and len(new_events) >= tail:
                    break
            else:
                break
        first = False
        if new_events:
            new_events = new_events[::-1]
            for x in new_events:
                dt = x['createdAt']
                dt = dt.astimezone(cst6cdt)
                dt = dt.strftime('%Y-%m-%d %H:%M:%S')
                print(f"[{dt}] {x['message']}")
        if primary and primary['rolloutState'] == 'COMPLETED':
            break
        # If the service ever stops returning new events you’ll just loop here
        time.sleep(interval)
    if primary:
        if primary['rolloutState'] == 'COMPLETED':
            notice_end(f'deployment completed successfully for service {service} in cluster {cluster}.')
        else:
            reason = primary['rolloutStateReason']
            notice_end(f'deployment failed: [{reason}]')



from raft.collection import Collection
from .subscription_tasks import subscriptions
from .resource_group_tasks import resource_groups
from .resource_group_tasks import providers
from .ldap_tasks import update_ldaps_cert
from .application_tasks import apps, register_app
from .group_tasks import groups
from .group_tasks import create_group
from .group_tasks import add_group_members
from .group_tasks import list_group_members
from .sharepoint_tasks import grant_access_to_site
from .sharepoint_tasks import revoke_site_access
from .sharepoint_tasks import list_site_permissions
from .whoami import whoami


azure_tasks = Collection()
azure_tasks.add_task(subscriptions)
azure_tasks.add_task(resource_groups)
azure_tasks.add_task(providers)
azure_tasks.add_task(update_ldaps_cert)
azure_tasks.add_task(apps)
azure_tasks.add_task(groups)
azure_tasks.add_task(create_group)
azure_tasks.add_task(add_group_members)
azure_tasks.add_task(list_group_members)
azure_tasks.add_task(grant_access_to_site)
azure_tasks.add_task(revoke_site_access)
azure_tasks.add_task(list_site_permissions)
azure_tasks.add_task(whoami)
azure_tasks.add_task(register_app)



from raft.tasks import task
from .base import AzureTask
from ..base.utils import dump_yaml
from ..base.utils import notice, notice_end


@task(klass=AzureTask)
def whoami(ctx, creds=None, **kwargs):
    """
    calls the whoami endpoint in the graph api
    """
    from affliction.graph_client import SynchronousGraphClient
    notice('calling whoami')
    tenant_id = kwargs.get('tenant_id')
    api = SynchronousGraphClient(tenant_id=tenant_id, creds=creds)
    result = { 'hello': 'sasame' }
    try:
        result = api.whoami()
        notice_end()
    except:
        notice_end()
        notice('appid')
        import jwt
        token = creds.get_token('https://graph.microsoft.com/.default')
        data = jwt.decode(
            token.token.encode('utf8'),
            options={'verify_signature': False})
        app_id = data['appid']
        notice_end(app_id)
        result = api.get(f"{api.base_url}/applications(appId='{app_id}')")
    dump_yaml(result, quiet=False)



from collections import defaultdict
from raft.tasks import task

from ...base.utils import dump_yaml
from ...base.utils import print_table
from ...base.utils import notice, notice_end
from ..base import AwsTask
from ..base import yielder


def get_records(session, zone_id):
    """
    Get all records in a zone
    """
    notice(f'getting records for {zone_id}')
    route53 = session.client('route53')
    kwargs = dict(
        HostedZoneId=zone_id,
        MaxItems='300',
    )
    rg = list(yielder(route53, 'list_resource_record_sets', **kwargs))
    records = defaultdict(dict)
    for record in rg:
        record_type = record['Type']
        name = record['Name']
        if 'ResourceRecords' in record:
            values = [ x['Value'] for x in record['ResourceRecords'] ]
            values.sort()
        else:
            values = record['AliasTarget']
        records[record_type][name] = (values, record)
    notice_end()
    return records


@task(klass=AwsTask)
def compare_zones(ctx, original_zone_id, original_profile,
                  new_zone_id, new_profile, cfn=False,
                  session=None, debug=False, **kwargs):
    """
    Compare two route53 zones
    """
    from boto3.session import Session
    original_session = Session(profile_name=original_profile)
    new_session = Session(profile_name=new_profile)
    original_records = get_records(original_session, original_zone_id)
    new_records = get_records(new_session, new_zone_id)
    header = [ 'type', 'name', 'status', 'values', 'new_values' ]
    rows = []
    for record_type, records in original_records.items():
        for name, t in records.items():
            values, record = t
            mp_new = new_records[record_type]
            if name not in mp_new:
                rows.append([record_type, name, 'missing', values, '' ])
                continue
            new_values, new_record = mp_new[name]
            if values != new_values:
                rows.append([record_type, name, 'different', values, new_values ])
    print_table(header, rows)
    if cfn:
        for row in rows:
            record_type = row[0]
            name = row[1]
            _, record = original_records[record_type][name]
            mp = {
                'Name': record['Name'],
                'HostedZoneId': '!Ref HostedZone',
                'Type': record['Type'],
            }
            if 'AliasTarget' in record:
                mp['AliasTarget'] = record['AliasTarget']
            elif 'ResourceRecords' in record:
                mp['ResourceRecords'] = [ x['Value'] for x in record['ResourceRecords'] ]
                mp['TTL'] = record['TTL']
            mp = { 'Type': 'AWS::Route53::RecordSet', 'Properties': mp, }
            pieces = [
                x.title().replace('-', '').replace('_', '')
                for x in name.split('.') ]
            pieces = pieces[:-1]
            if record_type == 'CNAME':
                pieces[0] = pieces[0].lower()
            mp = { 'Dot'.join(pieces): mp }
            dump_yaml(mp, quiet=False)



from raft import task

from convocations.aws.base import AwsTask


@task(klass=AwsTask)
def start_task(ctx, task_name, command='', cluster=None, tail=True, fargate=True, ec2=False, session=None, **kwargs):
    """
    kicks off a task in ecs

    ```yaml
    tundra:
      ecs:
        cluster: cluster1
        network_configuration:
          subnets:
            - subnet-0123456789
            - subnet-1234567890
          security_groups:
            - sg-0123456789
          assignPublicIp: DISABLED
    ```
    :param ctx:
    :param task_name: the name of the task
    :param command: the command to run inside the container
    :param cluster: the cluster name
    :param tail: if specified, we will tail the logs
    :param fargate: if specified, we will launch in fargate
    :param ec2: if specified, we will launch in ec2
    :param session:
    :param kwargs:

    :return:
    """
    from convocations.aws.ecs import get_latest_revision
    from convocations.aws.ecs import task_logs
    from convocations.base.utils import get_context_value, notice, notice_end
    x = get_latest_revision(ctx, task_name, session=session, quiet=True)
    ecs = session.client('ecs')
    if command:
        command = command.split(' ')
    cluster = cluster or get_context_value(ctx, 'tundra.ecs.cluster')
    network = get_context_value(ctx, 'tundra.ecs.network_configuration')
    if network:
        network = dict(awsvpcConfiguration=network)
    launch_type = 'FARGATE'
    if ec2:
        launch_type = 'EC2'
    notice(f'starting {x} on cluster {cluster} as {launch_type}')
    overrides = {
        'name': task_name,
    }
    if command:
        overrides['command'] = command
    params = dict(
        cluster=cluster,
        taskDefinition=x,
        overrides={
            'containerOverrides': [overrides]
        },
        launchType=launch_type,
    )
    if network:
        params['networkConfiguration'] = network
    result = ecs.run_task(**params)
    result = result['tasks'][0]
    task_arn = result['taskArn']
    task_id = task_arn.rsplit('/', 1)[-1]
    notice_end(task_id)
    if tail:
        task_logs(ctx, cluster, task_name, task_id, session=session, **kwargs)


@task(klass=AwsTask)
def deploy(ctx, cluster=None, service=None, session=None, **kwargs):
    """
    deploys a service in ecs, forcing a new deployment

    :param ctx:
    :param cluster: the cluster name
    :param service: the service name
    :param (boto3.session.Session) session: the session passed in from `AwsTask`
    :param kwargs:

    :return:
    """
    from ..base.utils import get_context_value
    from ..base.utils import notice, notice_end
    from botocore.exceptions import ClientError
    ecs = session.client('ecs')
    cluster = cluster or get_context_value(ctx, 'tundra.ecs.cluster')
    service = service or get_context_value(ctx, 'tundra.ecs.service')
    notice(f'deploying service {service} in cluster {cluster}')
    try:
        response = ecs.update_service(
            cluster=cluster,
            service=service,
            forceNewDeployment=True,
        )
        notice_end(f"{response['service']['status']}")
    except ClientError as ex:
        code = ex.response['Error']['Code']
        notice_end(code)
        if code in ('ServiceNotFoundException', 'ServiceNotActiveException'):
            # Service does not exist → stop the loop
            return
        # Anything else coming from ECS → propagate or handle as needed
        raise



from raft.tasks import task
from ..base import AwsTask


@task(klass=AwsTask)
def redshift_clusters(ctx, name='', session=None, **kwargs):
    """
    describes all the redshift clusters available
    """
    from ...base.utils import print_table
    client = session.client('redshift')
    response = client.describe_clusters()
    clusters = response['Clusters']
    rows = []
    for cluster in clusters:
        rows.append([
            cluster['ClusterIdentifier'],
            cluster['Endpoint']['Address'],
        ])
    print_table([ 'name', 'host' ], rows)
